{$X+,V-,B-,I-}
Program IPXGET;

{$DEFINE noTRACE}

uses crt,dos;

Const PEP_PACKET_TYPE       =4;  { Packet Exchange Protocol   }
      IPX_NOT_INSTALLED               = 240; {nwIpx}
      SHORT_LIVED_SOCKET = FALSE;

Type TnetworkAddress=array[1..4] of byte; { hi-endian }
     TnodeAddress   =array[1..6] of byte; { Hi-endian }
     TinterNetworkAddress=record
                          net   :TnetworkAddress; {hi-lo}
                          node  :Tnodeaddress;    {hi-lo}
                          socket:word;            {lo-hi}
                          end;

Type TipxHeader=Record
                checksum         :word; { not used, set to $FFFF }
                length           :word; { total number of bytes }
                TransportControl :byte; { used by bridges: low 4 bits= hop count
}
                packetType       :byte; { ignored by IPX, used by higher level
                                          protocols only. $00=unknown packet
type}
                destination,
                source           :TinternetWorkAddress;
                                         { if dest.network equals 0; dest
                                           assumed on same network as sender }
                                         { if dest.node =$FFFFFFFFFFFF, packet
                                           will be sent to all nodes. }
                end;
    { Fields within IPX and SPX are high-low. Byte swapping will be done
      by the IPX functions, except network and node addresses. }

    Tfragment=record { address and size of buffer fragment. }
              Address:Pointer;
              Size:word;
              end;

    Tecb=record
         Linkaddress     :Pointer; { used by IPX itself }
         ESRaddress      :Pointer;
         InUseFlag       :Byte; { reset to $00 when request completed }
         CompletionCode  :Byte; { valid after InUseFlag becomes $00;
                                 completionCode=$00: packet sent/received. }
         SocketNumber    :word;
         IPXworkspace    :array[1..4] of byte;
         DriverWorkspace :array[1..12] of byte;
         Immediateaddress:Tnodeaddress; { 6 bytes }
         FragmentCount   :word;  { must be >0 }
         Fragment        :array[1..2] of Tfragment; { [1..FragmentCount] }
                          { The number of fragments is unlimited.
                            However, most applications use 1 or 2. }
         end;
    Tpecb=^Tecb;

{    TAESecb=:
Offset  Size    Description
 00h    DWORD   Link
 04h    DWORD   ESR address
 08h    BYTE    in use flag (see below)
 09h  5 BYTEs   AES workspace }

Type TpepHeader=Record
               IPXhdr       :TipxHeader;  { set packettype to $04 }
               TransactionID:Longint;
               clientType   :word;
               end;



Var ListenECB     :Tecb;       { used to listen for packets }
    ListenPepHdr  :TpepHeader;

    SendECB       :Tecb;       { used to send acknowledgements }
    SendPepHdr    :TpepHeader;

    IOsocket      :word;
    DataBuffer    :array[1..546] of byte;
    SendDataBuffer:byte;

    PacketReceived   :Boolean;
    LastTransactionID:LongInt;

Var Result:word; { unit errorcode variable }

    NewStack:array[1..8192] of word;  { !! used by ESR }
    StackBottom:word;                 { !! used by ESR }

Var IpxSpxCall:Procedure;

Procedure IpxSpxSystemCall(Var regs:registers); assembler;
{ This method of calling IPX/SPX is preferred by Novell. }
{ For what its' worth: this call is 48 bytes longer than the other one.. }
asm
   { check if IpxSpxCall known. If not, return error $FF in fake AL }
xor ah,ah
mov al,$FF
les di,IpxSpxCall
mov bx,es
cmp bx,$0000
je  @1
   { move fake regs registers to 'real' registers }
   { AX, BX, CX, DX, SI, DI, ES only. }
les di,regs
mov ax,es:[di+16]
push ax            { push new es }
mov ax,es:[di+12]
push ax            { push new di }
mov ax,es:[di]
mov bx,es:[di+2]
mov cx,es:[di+4]
mov dx,es:[di+6]
mov si,es:[di+10]
pop di
pop es
   { farr call to A7 interrup handler }
push bp
CALL IpxSpxCall
pop bp
@1: { move 'real' registers to fake regs registers }
push es
push di
les di,regs
mov es:[di],ax
mov es:[di+2],bx
mov es:[di+4],cx
mov es:[di+6],dx
mov es:[di+10],si
pop ax             { ax:= 'di' }
mov es:[di+12],ax
pop ax             { ax:= 'es' }
mov es:[di+16],ax
end;


{IPX: 00}
Function IPXOpenSocket(Var socket:word; permanentSocket:boolean):boolean;
Var regs:registers;
    reqForSocket:boolean;
begin
regs.bx:=$0000;
if permanentSocket
 then regs.al:=$FF
 else regs.al:=$00;
regs.dx:=swap(socket); {hi-lo}
reqForSocket:=(socket=$0000);

IpxSpxSystemCall(Regs);

result:=regs.al;
if reqForSocket
 then socket:=swap(regs.dx); {force lo-hi}
IPXopenSocket:=(result=0);
{ resultcodes: $00 successful; $FE Socket Table Is Full;
               $FF socket already open OR IPX not initilazed. }
end;

{IPX: 01}
Function IPXCloseSocket(socket:word):boolean;
Var regs:registers;
begin
regs.bx:=$01;
regs.dx:=swap(socket);
IpxSpxSystemCall(regs);
result:=regs.al;
if result<>$FF then result:=$00;
IPXCloseSocket:=(result=$00);
{ possible resultcodes: $00 Successful; $FF IPX not initialized }
end;



Procedure CheckError(err:boolean; errNbr:word);
begin
if err
 then begin
      CASE errNbr of
       $0100:writeln('IPX needs to be installed.');
       $0101:writeln('ERROR: Connection not established. A Timeout occured');
       $0102:writeln('ERROR: The transfer is aborted; A timeout occured.');
       $0108:writeln('Transfer aborted.');
       $0300:writeln('The supplied path doesn'' exist / no write rights in directory.');
       $0301:writeln('Error writing to file / no write rights in directory.');
       $10FE:writeln('Error opening socket: Socket Table Is Full.');
       $10FF:writeln('Error opening socket: Socket is already open.');
       else writeln('Unspecified error.');
      end; {case}
      IPXcloseSocket(IOsocket);
      halt(1);
      end;
end;

Function TimeOut(t1,t2:word;n:byte):boolean;
{ ticks t2 - ticks t1 > n seconds ? }
Var lt1,lt2:LongInt;
begin
lt2:=t2;
if t1>t2 then lt2:=lt2+$FFFF;
TimeOut:=(lt2-t1)>(n*18);
end;

Function IPXListenForPacket(Var Ecb:Tecb):Boolean;
{ socket must be opened, ECB (partly) filled. }
Var regs:Registers;
begin
regs.bx:=$0004;
regs.es:=seg(ecb);
regs.si:=ofs(ecb);
IpxSpxSystemCall(Regs);
result:=regs.al;
if result<>$FF
 then result:=$00;
IpxListenForPacket:=(result=$00);
{resultcodes: $00 Successful;
              $FF Listening Socket doesn't exist OR IPX not initialized }
end;

Function IPXrelinquishControl:boolean;
Var regs:Registers;
begin
regs.bx:=$000A;
IpxSpxSystemCall(Regs);
result:=regs.al;
if result<>$FF then result:=$00;
IpxrelinquishControl:=(result=$00);
{resultcodes: $00 Successful; $FF IPX not initialized }
end;

Procedure PEPSetupListenECB(ESRptr:Pointer;ReceiveSocket:word;
                            BufPtr:Pointer;BufSize:word;
                     {out:} Var PepHdr:TpepHeader; Var ecb:Tecb);
{ Clears IPXheader and ECB, sets values of the required fields within
  the ecb and IPX header. }
{ ECB: ESR adress field, socket number, fragment count, frag.descriptor fields }
begin
FillChar(ecb,SizeOf(Tecb),#0);
FillChar(pepHdr,SizeOF(TpepHeader),#0);
WITH ECB
 do begin
    if ESRptr<>NIL
     then ESRaddress:=ESRptr;
    Fragmentcount:=2;
    socketNumber:=swap(ReceiveSocket); {hi-lo}

    Fragment[1].Address:=@pepHdr;
    Fragment[2].Address:=BufPtr;
    Fragment[1].size:=SizeOf(Tpepheader);
    Fragment[2].size:=BufSize;
    end;
end;

{IPX: 02}
Function IPXGetLocalTarget(Address:TinternetworkAddress;
                           Var ImmAddr:TnodeAddress;
                           VAR ticks:Word               ):boolean;
{ Ticks = estimated transmission time, in number of ticks (1/18 sec) }
Var reqAddr:TinternetworkAddress;
    repNode:TnodeAddress;
    Regs   :registers;
begin
move(Address,reqAddr,10);
reqAddr.socket:=swap(Address.socket); {hi-lo}
With regs
 do begin
    bx:=$0002;
    es:=seg(reqAddr); si:=ofs(reqAddr); di:=ofs(repNode);
    IpxSpxSystemCall(regs);
    ticks:=regs.cx;
    result:=regs.al;
    if result=0
     then move(repNode,ImmAddr,6);
    end;
IPXGetLocalTarget:=(result=$00);
{ resultcodes: $00 Successful; $FA No path to destination node found;
               $FF IPX not initialized. }
end;

Procedure PEPsetupSendECB(ESRptr:pointer; SourceSocket:word;
                          DestAddr:TinterNetworkAddress;
                          BufPtr:pointer; BufSize:word;
                   {out:} Var PepHdr:TpepHeader; Var ecb:Tecb);
{ Clears IPXheader and ECB, sets values of the required fields within
  the ecb and IPX header. }
Var ImmAddr:TnodeAddress;
    Ticks:word;
begin
fillchar(pepHdr,SizeOf(TpepHeader),#0);
with pepHdr.IPXhdr
 do begin
    PacketType:=PEP_PACKET_TYPE;
    Move(DestAddr,Destination,10);
    destination.socket:=swap(DestAddr.socket); {hi-lo}
    end;
IPXGetLocalTarget(DestAddr,ImmAddr,Ticks);
fillchar(ecb,sizeOf(ecb),#0);
With ecb
 do begin
    if ESRptr<>NIL
     then ESRaddress:=ESRptr;
    socketNumber:=swap(SourceSocket); {hi-lo}
    Move(ImmAddr,ImmediateAddress,6);
    FragmentCount:=2;
    fragment[1].Address:=@pephdr;
    fragment[1].size:=SizeOf(TpepHeader);
    fragment[2].Address:=BufPtr;
    fragment[2].size:=BufSize;
    end;
end;
Function IPXSendPacket(Var Ecb:Tecb):boolean;
{ the ecb must be filled, before calling this function }
{ Right after this call, IPXrelinquishControl should be called Iteratively,
  this allows the sending of the IPX packet. }
Var regs:Registers;
begin
regs.bx:=$0003;
regs.es:=seg(ecb);
regs.si:=ofs(ecb);
IpxSpxSystemCall(Regs);
result:=regs.al;
if result<>$FF then result:=$00;
IpxSendPacket:=(result=$00);
{ possible resultcodes: $00 Successful; $FF IPX not initialized }
end;


{$F+}
Procedure ListenAndAckHandler;
begin
If (ListenECB.CompletionCode<>0)
 or (ListenPepHdr.IPXhdr.packetType<>PEP_PACKET_TYPE)
 or (ListenPepHdr.clienttype<>$EA)
 or (ListenPepHdr.TransactionID<LastTransactionID) { discard dupe old packet }
  then IPXlistenForPacket(ListenECB)
  else begin
       PacketReceived:=(ListenPepHdr.transactionID>LastTransactionID); { new
packet received }

       { Acknowledge new packets and duplicates of the latest packet, }
       { as the original acknowledgement may have been lost. }
       LastTransactionID:=ListenPepHdr.TransactionID;

       { Setup acknowledgement ECB and PEPheader, and send it. }
       if SendECB.InUseFlag=0
        then begin

ListenPepHdr.IPXhdr.source.socket:=swap(ListenPepHdr.IPXhdr.source.socket);
             { socket is hi-lo in IPX/PEPHeaders. SetupSendECB expects lo-hi }
             PEPsetupSendECB(NIL,IOsocket,ListenPepHdr.IPXhdr.source,
                             @SendDataBuffer,0,
                             SendPepHdr,SendECB);
             SendPepHdr.TransactionId:=LastTransactionID;
             SendPepHdr.ClientType:=$EA;
             IPXsendPacket(SendECB);
             end;
       end;
end;
{$F-}

{$F+}
Procedure ListenAndAckESR; assembler;
asm
    mov dx, seg stackbottom
    mov ds, dx

    mov dx,ss  { setup of a new local stack }
    mov bx,sp  { ss:sp copied to dx:bx}
    mov ax,ds
    mov ss,ax
    mov sp,offset stackbottom
    push dx    { push old ss:sp on new stack }
    push bx
    CALL ListenAndAckHandler
    pop bx     { restore ss:sp from new stack }
    pop dx
    mov sp,bx
    mov ss,dx
end;
{$F-}

Function IPXinitialize:Boolean;
CONST DOS_MULTIPLEX =$2F;
Var regs:registers;
begin
Regs.AX:=$7A00;
INTR(DOS_MULTIPLEX,Regs);
if regs.AL<>$FF
 then begin
      Result:=IPX_NOT_INSTALLED;
      IpxInitialize:=false
      end
 else begin
      @IpxSpxCall:=Ptr(Regs.es,Regs.di);
      Result:=0;
      IpxInitialize:=true;
      end;
end;

Function IPXgetIntervalMarker(Var ticks:word):boolean;
Var regs:Registers;
begin
regs.bx:=$0008;
IpxSpxSystemCall(Regs);
ticks:=regs.ax;
result:=$00;
IPXgetIntervalMarker:=True;
end;

Var ticks,ticks2 :word;
    FileName:string;
    FileSize:LongInt;
    DirName:string;
    f:file;
    BytesToWrite,BytesWritten:word;

begin
IpxInitialize;
CheckError(result>0,$100);

If (pos('?',ParamStr(1))>0) or (paramcount>1)
 then begin
      writeln;
      writeln('Usage: IPXGET <directory>');
      writeln;
      writeln('The file sent by IPXSEND on another workstation');
      writeln('will be copied to the supplied directory.');
      halt(1);
      end;

If paramcount=1
 then DirName:=ParamStr(1)
 else DirName:='.';

IF NOT (DirName[ord(dirName[0])] IN [':','\'])
 then DirName:=DirName+'\';
assign(f,DirName+'temp.$$$');
rewrite(f,1);
CheckError(IOresult<>0,$0300);
close(f);

IOSocket:=$5678;
IPXopenSocket(IOsocket,SHORT_LIVED_SOCKET);
CheckError(result>0,$1000+result);

{ Setup of ECB and PepHeader, start listening for incoming packets. }
LastTransactionID:=0;
PacketReceived:=False;
PEPSetupListenECB(Addr(ListenAndAckESR),IOsocket,@DataBuffer,546,
                  ListenPepHdr,ListenECB);
IPXListenForPacket(ListenECB);
writeln('Waiting for IPXSEND to start sending.. (any key to abort)');

IPXGetIntervalMarker(ticks);
REPEAT
 IPXrelinquishControl;
 IPXGetIntervalMarker(ticks2);
 CheckError(TimeOut(ticks,ticks2,130),$101);{ error if a timeout occurred }
 CheckError(Keypressed,$108);
UNTIL PacketReceived;

writeln('Handshaking.. Initiating transfer process.');
{$IFDEF TRACE}
writeln('Received PacketID:',LastTransactionID);
{$ENDIF}

{ do something with DataBuffer: the data that was just received. }
{ the first packet contains the filename and filesize }
Move(DataBuffer[1],FileName[0],15);
Move(DataBuffer[16],FileSize,4);
writeln('Receiving file ',FileName,', size: ',FileSize);

assign(f,DirName+filename);
rewrite(f,1);
BytesToWrite:=512;

REPEAT { Listen for remaining packets  }
 Packetreceived:=false;

 While SendECB.InuseFlag<>0
  do IPXrelinquishControl;

 IPXListenForPacket(ListenECB);
 IPXGetIntervalMarker(ticks);
 Repeat
  IPXrelinquishControl;
  IPXGetIntervalMarker(ticks2);
  CheckError(TimeOut(ticks,ticks2,10),$102); { error if Timeout occurred }
  CheckError(Keypressed,$108);
 until PacketReceived;
 {$IFDEF TRACE}
 writeln('Received packet#:',LastTransactionID);
 {$ENDIF}

 { Write DataBuffer to disk. }
 IF FileSize<512
  then BytesToWrite:=FileSize;
 BlockWrite(f,DataBuffer,BytesToWrite,BytesWritten);
 CheckError(BytesToWrite<>BytesWritten,$0301);

 FileSize:=FileSize-512;
UNTIL (FileSize<=0); { entire file received }

writeln('Transfer complete.');
IPXcloseSocket(IOsocket);
close(f);
end.
